----------------------------------------------------------------------------------
-- Company: NTNU 
-- Engineer: Martin Lesund
-- 
-- Create Date: 03/23/2017 03:19:50 PM
-- Design Name: 
-- Module Name: SPI_master - Behavioral
-- Project Name: Master thesis
-- Target Devices: 
-- Tool Versions: vivado 2016.3
-- Description: This is the master file for the SPI
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
----------------------------------------------------------------------------------
--PARAMETER description:
-- g_ = generic
-- i_ = input
-- o_ = output
-- r_ = register signal (has registered logic)
-- s_ = State Machine state
----------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


entity SPI_master is
  generic (
    g_data_bus_width : integer := 16;   -- Data bus width   
    g_num_of_slaves  : integer := 1 ;   -- Number of SPI slaves
    g_divide_value   : integer := 4 ;   --input o_SPI_sclk freq = i_clock/(2*g_divide_value)
    -------CPOL and CPHA MODES -----------------------------
    g_CPOL             : std_logic := '1';
    g_CPHA             : std_logic := '1'
      );
  Port 
  ( 
    -- USER LOGIC -------------------------------------------------------------------------- 
    i_clock        : in     std_logic;                                    -- System clock (50Mhz)
    i_reset_n      : in     std_logic;                                    -- Async reset (active high)
    ----------------------------------------------------------------------------------------
    i_enable      : in     std_logic;                                      -- Enable signal
    i_write_data  : in     std_logic_vector(g_data_bus_width-1 downto 0);  -- Transmitted Data to slaves
    o_done        : out    std_logic;                                      -- Transmission done signal
    o_read_data   : out    std_logic_vector(g_data_bus_width-1 downto 0);  -- Received data from slave
    i_addr        : in     std_logic_vector(g_num_of_slaves-1 downto 0);   -- Address of slave
    
    -- SPI MASTER INTERFACE -----------------------------------------------------------------
    o_SPI_sclk    : out    std_logic;                                    -- SPI clock
   	o_SPI_ss_n    : out    std_logic_vector(g_num_of_slaves-1 downto 0); -- Slave select
    o_SPI_mosi    : out    std_logic;                                    -- SPI master out, slave in, serial data line
    i_SPI_miso    : in     std_logic                                     -- SPI master in, slave out, serial data line 
  

  );
end SPI_master;

architecture logic of SPI_master is
    type state is (s_idle, s_transmission, s_transmission_end);
    signal present_state, next_state : state;
    attribute enum_encoding: string;                          
    -- Optional attribute  IMPORTANT FOR OPTIMAL USAGE (sequential vs. Gray vs. One-Hot)
    attribute enum_encoding of state: type is "00 01 11";   
    -- Describes which encoding style used
   
   --Clock signals
   signal r_rise_sclk       : std_logic;
   signal r_fall_sclk       : std_logic;
   signal r_counter_enable  : std_logic;
   signal r_counter         : integer range 0 to g_divide_value*2;
   signal r_counter_data    : integer range 0 to g_data_bus_width;
   signal w_counter_data    : std_logic;
   signal r_transmit_start   : std_logic;  
   signal r_tx_data          : std_logic_vector(g_data_bus_width-1 downto 0);
   signal r_rx_data          : std_logic_vector(g_data_bus_width-1 downto 0);
   signal r_slave             : integer := 0;         -- slave selected for current transaction

begin

w_counter_data <= '0' when(r_counter_data>0) else '1';
  
 -------------------------------------------------------------------------------
-- Counter Clock generator
--------------------------------------------------------------------------------

p_spi_clock_gen : process (i_clock, i_reset_n) 
begin
    if(i_reset_n = '0') then
        r_rise_sclk <= '0';
        r_fall_sclk <= '0';
        r_counter   <= 0;    
     elsif (rising_edge (i_clock)) then
        if (r_counter_enable = '1' and   g_CPOL = '1') then   -- r_Mode11
            if (r_counter = g_divide_value-1) then  -- first edge  = fall
                r_rise_sclk <= '0';
                r_fall_sclk <= '1';
                r_counter   <= r_counter + 1;
            elsif(r_counter = (g_divide_value*2)-1) then 
                r_rise_sclk <= '1';
                r_fall_sclk <= '0';
                r_counter   <= 0;
            else
                r_rise_sclk <= '0';
                r_fall_sclk <= '0';
                r_counter   <= r_counter+1;
            end if;
         elsif (r_counter_enable = '1' and g_CPOL = '0' ) then   -- r_Mode01                  
                        if (r_counter = g_divide_value-1) then  --first edge = rise
                            r_rise_sclk <= '1';
                            r_fall_sclk <= '0';
                            r_counter   <= r_counter + 1;
                        elsif(r_counter = (g_divide_value*2)-1) then
                            r_rise_sclk <= '0';
                            r_fall_sclk <= '1';
                            r_counter   <= 0;
                        else
                            r_rise_sclk <= '0';
                            r_fall_sclk <= '0';
                            r_counter   <= r_counter+1;
                        end if;                                    
         else
           r_rise_sclk <= '0';
           r_fall_sclk <= '0';
           r_counter   <=  0;
         end if;
       end if;
end process p_spi_clock_gen;


----------------------------------------------------------------------------------
-- FSM 
----------------------------------------------------------------------------------

  ------------ LOWER SECTION OF FSM: ------------ (Sequential Circuit)
  p_fsm_present_state : process ( i_clock, i_reset_n)
  begin
        if (i_reset_n = '0') then
            present_state <= s_idle;
        elsif(rising_edge(i_clock)) then
         present_state <= next_state;
        end if;    
  end process p_fsm_present_state;

  ------------ UPPER SECTION OF FSM: ------------ (Combinational Circuit)
  p_fsm_next_state: process (present_state, r_rise_sclk, r_fall_sclk, r_transmit_start,w_counter_data)
  begin  
   case present_state is
    when s_transmission =>
        if ((r_rise_sclk = '1' and g_CPOL = '1') or (r_fall_sclk ='1' and g_CPOL ='0')) and (w_counter_data = '1') then  
            next_state <= s_transmission_end;
        else
            next_state <= s_transmission;
        end if;
    when s_transmission_end =>
        if (r_fall_sclk ='1' and g_CPOL = '1') or (r_rise_sclk = '1' and g_CPOL ='0') then 
            next_state <= s_idle;
        else
            next_state <= s_transmission_end;
        end if;
    when others =>                                    --s_IDLE
        if r_transmit_start = '1' then 
            next_state <= s_transmission;
        else
            next_state <= s_idle;
        end if;
   end case;   
  end process p_fsm_next_state;

  ------------ OUTPUT SECTION: ------------ 
  p_fsm_outputs : process (i_clock, i_reset_n)             
  begin
    if(i_reset_n = '0') then
        r_transmit_start <= '0';
        r_tx_data       <= (others => '0');
        r_rx_data       <= (others => '0');
        o_done        <= '0';
        o_read_data       <= (others => '0');
        
        r_counter_data <= g_data_bus_width -1;
        r_counter_enable <= '0';
        o_SPI_sclk          <= g_CPOL;                   -- Clock polarity 
        o_SPI_ss_n         <= (others => '1');         -- Set all slave select outputs high
        o_SPI_mosi          <= '1';
     elsif (rising_edge(i_clock)) then
        r_transmit_start <= i_enable;
		          
        
    
      case present_state is 
          when s_transmission =>
            r_counter_enable    <= '1';
            o_done            <= '0';
 ----------------------------------------------MODE11----------------------------------------------------               
            if(r_rise_sclk = '1' and g_CPOL = '1') then                    
                o_SPI_sclk          <=  '1'; 
                r_rx_data       <= r_rx_data(g_data_bus_width-2 downto 0)&i_SPI_miso; --left shift
                 if(r_counter_data>0) then
                    r_counter_data <= r_counter_data -1;
                 end if;
            elsif(r_fall_sclk = '1'and g_CPOL = '1')then                      
                o_SPI_sclk          <= '0' ; 
                o_SPI_mosi          <= r_tx_data(g_data_bus_width-1);
                r_tx_data       <= r_tx_data(g_data_bus_width-2 downto 0)&'1';
			
  ----------------------------------------------MODE01----------------------------------------------------              
            elsif(r_fall_sclk = '1' and g_CPOL ='0')then                      
                 o_SPI_sclk          <= '0' ; 
                 r_rx_data       <= r_rx_data(g_data_bus_width-2 downto 0)&i_SPI_miso; --left shift
                     if(r_counter_data>0) then
                        r_counter_data <= r_counter_data -1;
                     end if;

  
            elsif(r_rise_sclk ='1' and g_CPOL ='0') then                    
                 o_SPI_sclk          <=  '1'; 
                  o_SPI_mosi          <= r_tx_data(g_data_bus_width-1);
                  r_tx_data       <= r_tx_data(g_data_bus_width-2 downto 0)&'1';
			end if;	                                                    
            
            o_SPI_ss_n(r_slave) <= '0';
     
          when s_transmission_end =>
           if (g_CPOL = '1' ) then              
             o_done            <= r_fall_sclk;             
             o_SPI_ss_n(r_slave)     <= '0';                                              
             o_read_data           <= r_rx_data;
             r_counter_data      <= g_data_bus_width-1;
             r_counter_enable    <= '1';
            elsif(g_CPOL = '0') then
             o_done            <= r_rise_sclk;            
             o_SPI_ss_n(r_slave)     <= '0';                                             
             o_read_data           <= r_rx_data;
             r_counter_data      <= g_data_bus_width-1;
             r_counter_enable    <= '1';
            
            end if;

           
          --s_idle
          when others => 
               o_SPI_sclk <= g_CPOL;                       --set spi clock polarity   
               o_SPI_ss_n<= (others => '1');              -- Set all slaves to '1'
               o_SPI_mosi <= '1';
               r_tx_data <= i_write_data;    
               o_done <= '0';
               r_counter_data <= g_data_bus_width-1;
               r_counter_enable <= '0';
            if(to_integer(unsigned(i_addr)) < g_num_of_slaves) then
               r_slave <= to_integer(unsigned(i_addr));   -- set current slave selection if valid
            else 
               r_slave <= 0;                              -- set to first slave if not valid
            end if;             
       

      end case ;
     end if;
  end process p_fsm_outputs;

end logic;
